/* Copyright (C) 2010-2021 Frankfurt Institute for Advanced Studies, Goethe-Universität Frankfurt, Frankfurt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Igor Kulakov [committer], Maksym Zyzak, Valentina Akishina */

#include "CaTripletConstructor.h"

#include "CaDefines.h"
#include "CaFramework.h"
#include "CaGridArea.h"

#include <algorithm>
#include <iostream>
// #include "CaToolsDebugger.h"
#include "AlgoFairloggerCompat.h"
#include "CaTrackFit.h"
#include "CaTrackParam.h"

// using cbm::ca::tools::Debugger;

using namespace cbm::algo::ca;

TripletConstructor::TripletConstructor(ca::Framework& algo, WindowData& wData) : frAlgo(algo), frWData(wData) {}

void TripletConstructor::InitStations(int istal, int istam, int istar)
{
  fIsInitialized = false;

  fIstaL = istal;
  fIstaM = istam;
  fIstaR = istar;

  if (fIstaM >= frAlgo.fParameters.GetNstationsActive()) {
    return;
  }

  fStaL = &frAlgo.fParameters.GetStation(fIstaL);
  fStaM = &frAlgo.fParameters.GetStation(fIstaM);
  fStaR = &frAlgo.fParameters.GetStation(fIstaR);


  {  // two stations for approximating the field between the target and the left hit
    int sta1 = fIstaL;
    if (sta1 >= frAlgo.fNfieldStations) {
      sta1 = frAlgo.fNfieldStations - 1;
    }
    if (sta1 < 1) {
      sta1 = 1;
    }
    int sta0 = sta1 / 2;  // station between fIstaL and the target

    assert(0 <= sta0 && sta0 < sta1 && sta1 < frAlgo.fParameters.GetNstationsActive());

    fFld0Sta[0] = &frAlgo.fParameters.GetStation(sta0);
    fFld0Sta[1] = &frAlgo.fParameters.GetStation(sta1);
  }

  {  // three stations for approximating the field between the left and the right hit

    int sta0 = fIstaL;
    int sta1 = fIstaM;
    int sta2 = fIstaM + 1;
    if (sta2 >= frAlgo.fParameters.GetNstationsActive()) {
      sta2 = fIstaM;
      sta1 = fIstaM - 1;
      if (sta0 >= sta1) {
        sta0 = sta1 - 1;
      }
      if (sta0 < 0) {
        sta0 = 0;
      }
    }
    if (frAlgo.fParameters.GetNstationsActive() >= 3) {
      assert(0 <= sta0 && sta0 < sta1 && sta1 < sta2 && sta2 < frAlgo.fParameters.GetNstationsActive());
    }

    fFld1Sta[0] = &frAlgo.fParameters.GetStation(sta0);
    fFld1Sta[1] = &frAlgo.fParameters.GetStation(sta1);
    fFld1Sta[2] = &frAlgo.fParameters.GetStation(sta2);
  }

  fFit.SetParticleMass(frAlgo.GetDefaultParticleMass());
  if (frWData.CurrentIteration()->GetElectronFlag()) {
    fFit.SetParticleMass(constants::phys::ElectronMass);
  }
  fFit.SetMask(fmask::One());
  fFit.SetDoFitVelocity(true);

  fIsInitialized = true;
}


void TripletConstructor::CreateTripletsForHit(int istal, int istam, int istar, ca::HitIndex_t iHitL)
{
  InitStations(istal, istam, istar);

  if (!fIsInitialized) return;

  fIhitL        = iHitL;
  const auto& hitl = frWData.Hit(fIhitL);

  // fit a straight line through the target and the left hit.
  {

    /// Get the field approximation. Add the target to parameters estimation.
    /// Propagaete to the middle station of a triplet.

    fFit.SetQp0(frWData.CurrentIteration()->GetMaxQp());

    ca::FieldValue<fvec> lB, mB, cB, rB _fvecalignment;
    ca::FieldValue<fvec> l_B_global, centB_global _fvecalignment;

    // get the magnetic field along the track.
    // (suppose track is straight line with origin in the target)

    fvec xl         = hitl.X();
    fvec yl         = hitl.Y();
    fvec zl         = hitl.Z();
    fvec time       = hitl.T();
    fvec timeEr2    = hitl.dT2();
    const fvec dzli = 1.f / (zl - frAlgo.GetParameters().GetTargetPositionZ());

    const fvec tx = (xl - frAlgo.GetParameters().GetTargetPositionX()) * dzli;
    const fvec ty = (yl - frAlgo.GetParameters().GetTargetPositionY()) * dzli;

    TrackParamV& T = fFit.Tr();

    T.X()    = xl;
    T.Y()    = yl;
    T.Z()    = zl;
    T.Tx()   = tx;
    T.Ty()   = ty;
    T.Qp()   = fvec(0.);
    T.Time() = time;
    T.Vi()   = constants::phys::SpeedOfLightInv;

    const fvec maxSlopePV = frWData.CurrentIteration()->GetMaxSlopePV();
    const fvec maxQp      = frWData.CurrentIteration()->GetMaxQp();
    fvec txErr2           = maxSlopePV * maxSlopePV / fvec(9.);
    fvec qpErr2           = maxQp * maxQp / fvec(9.);

    T.ResetErrors(1., 1., txErr2, txErr2, qpErr2, (staL().timeInfo ? timeEr2 : 1.e6), 1.e10);

    T.InitVelocityRange(1. / frWData.CurrentIteration()->GetMaxQp());

    T.C00() = hitl.dX2();
    T.C10() = hitl.dXY();
    T.C11() = hitl.dY2();

    T.Ndf()     = (frWData.CurrentIteration()->GetPrimaryFlag()) ? fvec(2.) : fvec(0.);
    T.NdfTime() = (staL().timeInfo ? 0 : -1);

    // NDF = number of track parameters (6: x, y, tx, ty, qp, time) - number of measured parameters (3: x, y, time) on station or (2: x, y) on target

    // field made by  the left hit, the target and the station istac in-between.
    // is used for extrapolation to the target and to the middle hit
    ca::FieldRegion<fvec> fld0;
    {
      ca::FieldValue<fvec> B0 = fFld0Sta[0]->fieldSlice.GetFieldValueForLine(T);
      ca::FieldValue<fvec> B1 = fFld0Sta[1]->fieldSlice.GetFieldValueForLine(T);
      fld0.Set(frWData.TargB(), frAlgo.GetParameters().GetTargetPositionZ(), B0, fFld0Sta[0]->fZ, B1, fFld0Sta[1]->fZ);
    }

    {  // field, made by the left hit, the middle station and the right station
      // Will be used for extrapolation to the right hit
      ca::FieldValue<fvec> B0 = fFld1Sta[0]->fieldSlice.GetFieldValueForLine(T);
      ca::FieldValue<fvec> B1 = fFld1Sta[1]->fieldSlice.GetFieldValueForLine(T);
      ca::FieldValue<fvec> B2 = fFld1Sta[2]->fieldSlice.GetFieldValueForLine(T);
      fFldL.Set(B0, fFld1Sta[0]->fZ, B1, fFld1Sta[1]->fZ, B2, fFld1Sta[2]->fZ);
    }

    // add the target constraint

    fFit.FilterWithTargetAtLine(frAlgo.GetParameters().GetTargetPositionZ(), frWData.TargetMeasurement(), fld0);

    fFit.MultipleScattering(frAlgo.fParameters.GetMaterialThickness(fIstaL, T.GetX(), T.GetY()));

    // extrapolate to the middle hit

    fFit.ExtrapolateLine(staM().fZ, fFldL);

    fTrL = T;
  }

  FindDoublets();
  FitDoublets();
  FindTriplets();
}


void TripletConstructor::FindDoublets()
{
  /// Find the doublets. Reformat data into portions of doublets.

  int iMC = -1;
  if (frAlgo.fParameters.DevIsMatchDoubletsViaMc()) {
    iMC = frAlgo.GetMcTrackIdForWindowHit(fIhitL);
    if (iMC < 0) {
      return;
    }
  }

  CollectHits(fTrL, fIstaM, frWData.CurrentIteration()->GetDoubletChi2Cut(), iMC, fHitsM_2,
              frAlgo.fParameters.GetMaxDoubletsPerSinglet());
}


void TripletConstructor::FitDoublets()
{

  // ---- Add the middle hits to parameters estimation ----

  Vector<ca::HitIndex_t> hitsMtmp("TripletConstructor::hitsMtmp", fHitsM_2);

  fHitsM_2.clear();
  fTracks_2.clear();
  fTracks_2.reserve(hitsMtmp.size());

  bool isMomentumFitted = (frAlgo.fIsTargetField || (staL().fieldStatus != 0) || (staM().fieldStatus != 0));

  double maxQp = frWData.CurrentIteration()->GetMaxQp();
  for (unsigned int i2 = 0; i2 < hitsMtmp.size(); i2++) {

    const ca::HitIndex_t indM = hitsMtmp[i2];
    const ca::Hit& hitm       = frWData.Hit(indM);

    if (frWData.IsHitSuppressed(indM)) {
      continue;
    }

    TrackParamV& T2 = fFit.Tr();
    T2              = fTrL;
    fFit.SetQp0(fvec(0.f));

    fvec z_2 = hitm.Z();
    ca::MeasurementXy<fvec> m_2(hitm.X(), hitm.Y(), hitm.dX2(), hitm.dY2(), hitm.dXY(), fvec::One(), fvec::One());
    fvec t_2   = hitm.T();
    fvec dt2_2 = hitm.dT2();

    // add the middle hit

    fFit.ExtrapolateLineNoField(z_2);

    fFit.FilterXY(m_2);

    fFit.FilterTime(t_2, dt2_2, staM().timeInfo);

    fFit.SetQp0(isMomentumFitted ? fFit.Tr().GetQp() : maxQp);

    fFit.MultipleScattering(frAlgo.fParameters.GetMaterialThickness(fIstaM, T2.GetX(), T2.Y()));

    fFit.SetQp0(fFit.Tr().Qp());

    // check if there are other hits close to the doublet on the same station
    if (ca::Framework::kMcbm != frAlgo.fTrackingMode) {
      // TODO: SG: adjust cuts, put them to parameters

      fscal tx = T2.Tx()[0];
      fscal ty = T2.Ty()[0];
      fscal tt = T2.Vi()[0] * sqrt(1. + tx * tx + ty * ty);  // dt/dl * dl/dz

      for (unsigned int iClone = i2 + 1; iClone < hitsMtmp.size(); iClone++) {

        int indClone            = hitsMtmp[iClone];
        const ca::Hit& hitClone = frWData.Hit(indClone);

        fscal dz = hitClone.Z() - T2.Z()[0];

        if ((staM().timeInfo) && (T2.NdfTime()[0] >= 0)) {
          fscal dt = T2.Time()[0] + tt * dz - hitClone.T();
          if (!(fabs(dt) <= 3.5 * sqrt(T2.C55()[0]) + hitClone.RangeT())) {
            continue;
          }
        }

        fscal dx = T2.GetX()[0] + tx * dz - hitClone.X();
        if (!(fabs(dx) <= 3.5 * sqrt(T2.C00()[0]) + hitClone.RangeX())) {
          continue;
        }

        fscal dy = T2.Y()[0] + ty * dz - hitClone.Y();
        if (!(fabs(dy) <= 3.5 * sqrt(T2.C11()[0]) + hitClone.RangeY())) {
          continue;
        }

        if (frAlgo.fParameters.DevIsSuppressOverlapHitsViaMc()) {
          int iMC = frAlgo.GetMcTrackIdForWindowHit(fIhitL);
          if ((iMC != frAlgo.GetMcTrackIdForWindowHit(indM)) || (iMC != frAlgo.GetMcTrackIdForWindowHit(indClone))) {
            continue;
          }
        }
        frWData.SuppressHit(indClone);
      }
    }

    fHitsM_2.push_back(indM);
    fTracks_2.push_back(T2);
  }  // i2
}


void TripletConstructor::FindTriplets()
{
  if (fIstaR >= frAlgo.fParameters.GetNstationsActive()) {
    return;
  }
  FindRightHit();
  FitTriplets();
  StoreTriplets();
}


void TripletConstructor::FindRightHit()
{

  /// Add the middle hits to parameters estimation. Propagate to right station.
  /// Find the triplets(right hit). Reformat data in the portion of triplets.

  fFit.SetQp0(fvec(0.));

  if (fIstaR >= frAlgo.fParameters.GetNstationsActive()) return;

  {
    int maxTriplets = fHitsM_2.size() * frAlgo.fParameters.GetMaxTripletPerDoublets();
    fHitsM_3.clear();
    fHitsR_3.clear();
    fHitsM_3.reserve(maxTriplets);
    fHitsR_3.reserve(maxTriplets);
  }
  // ---- Add the middle hits to parameters estimation. Propagate to right station. ----

  double maxSlope       = frWData.CurrentIteration()->GetMaxSlope();
  double tripletChi2Cut = frWData.CurrentIteration()->GetTripletChi2Cut();
  for (unsigned int i2 = 0; i2 < fHitsM_2.size(); i2++) {

    fFit.SetTrack(fTracks_2[i2]);
    TrackParamV& T2 = fFit.Tr();

    // extrapolate to the right hit station

    fFit.Extrapolate(staR().fZ, fFldL);

    if constexpr (fDebugDublets) {
      ca::HitIndex_t iwhit[2] = {fIhitL, fHitsM_2[i2]};
      ca::HitIndex_t ihit[2]  = {frWData.Hit(iwhit[0]).Id(), frWData.Hit(iwhit[1]).Id()};

      int ih0           = ihit[0];
      int ih1           = ihit[1];
      const ca::Hit& h0 = frAlgo.fInputData.GetHit(ih0);
      const ca::Hit& h1 = frAlgo.fInputData.GetHit(ih1);

      LOG(info) << "\n======  Extrapolated Doublet : "
                << "  iter " << frAlgo.fCurrentIterationIndex << " hits: {" << fIstaL << "/" << ih0 << " " << fIstaM
                << "/" << ih1 << " " << fIstaR << "/?} xyz: {" << h0.X() << " " << h0.Y() << " " << h0.Z() << "}, {"
                << h1.X() << " " << h1.Y() << " " << h1.Z() << "} chi2 " << T2.GetChiSq()[0] << " ndf " << T2.Ndf()[0]
                << " chi2time " << T2.ChiSqTime()[0] << " ndfTime " << T2.NdfTime()[0];
      LOG(info) << "  extr. track: " << T2.ToString(0);
    }

    // ---- Find the triplets(right hit). Reformat data in the portion of triplets. ----
    bool isDoubletGood = false;
    int iMC            = -1;
    do {
      if (frAlgo.kSts == frAlgo.fTrackingMode && (T2.C44()[0] < 0)) {
        break;
      }
      if (T2.C00()[0] < 0 || T2.C11()[0] < 0 || T2.C22()[0] < 0 || T2.C33()[0] < 0 || T2.C55()[0] < 0) {
        break;
      }
      if (fabs(T2.Tx()[0]) > maxSlope) {
        break;
      }
      if (fabs(T2.Ty()[0]) > maxSlope) {
        break;
      }
      if (frAlgo.fParameters.DevIsMatchTripletsViaMc()) {
        int indM = fHitsM_2[i2];
        iMC      = frAlgo.GetMcTrackIdForWindowHit(fIhitL);
        if (iMC < 0 || iMC != frAlgo.GetMcTrackIdForWindowHit(indM)) {
          break;
        }
      }
      isDoubletGood = true;
    } while (0);

    if constexpr (fDebugDublets) {
      if (isDoubletGood) {
        LOG(info) << "  extrapolated doublet accepted";
      }
      else {
        LOG(info) << "  extrapolated doublet rejected";
        LOG(info) << "======== end of extrapolated doublet ==== \n";
      }
    }

    if (!isDoubletGood) {
      continue;
    }

    Vector<ca::HitIndex_t> collectedHits;
    CollectHits(T2, fIstaR, tripletChi2Cut, iMC, collectedHits, frAlgo.fParameters.GetMaxTripletPerDoublets());

    if (collectedHits.size() >= frAlgo.fParameters.GetMaxTripletPerDoublets()) {
      LOG(debug) << "Ca: GetMaxTripletPerDoublets==" << frAlgo.fParameters.GetMaxTripletPerDoublets()
                 << " reached in findTripletsStep0()";
      // reject already created triplets for this doublet
      collectedHits.clear();
    }
    if constexpr (fDebugDublets) {
      LOG(info) << " collected " << collectedHits.size() << " hits on the right station ";
    }
    for (unsigned int ih = 0; ih < collectedHits.size(); ih++) {
      ca::HitIndex_t irh = collectedHits[ih];
      if constexpr (fDebugDublets) {
        ca::HitIndex_t ihit = frWData.Hit(irh).Id();
        const ca::Hit& h    = frAlgo.fInputData.GetHit(ihit);
        LOG(info) << "  hit " << ihit << " " << h.ToString();
      }
      if (frWData.IsHitSuppressed(irh)) {
        if constexpr (fDebugDublets) {
          LOG(info) << "  the hit is suppressed";
        }
        continue;
      }
      // pack the triplet
      fHitsM_3.push_back(fHitsM_2[i2]);
      fHitsR_3.push_back(irh);
    }  // search area
    if constexpr (fDebugDublets) {
      LOG(info) << "======== end of extrapolated doublet ==== \n";
    }
  }  // i2
}


void TripletConstructor::FitTriplets()
{
  constexpr int nIterations = 2;

  Tindex n3 = fHitsM_3.size();
  assert(fHitsM_3.size() == fHitsR_3.size());

  fTracks_3.clear();
  fTracks_3.reset(n3, TrackParamV());

  /// Refit Triplets
  if constexpr (fDebugTriplets) {
    //cbm::ca::tools::Debugger::Instance().AddNtuple(
    //  "triplets", "ev:iter:i0:x0:y0:z0:i1:x1:y1:z1:i2:x2:y2:z2:mc:sta:p:vx:vy:vz:chi2:ndf:chi2time:ndfTime");
  }

  ca::TrackFit fit;
  fit.SetMask(fmask::One());

  if (frWData.CurrentIteration()->GetElectronFlag()) {
    fit.SetParticleMass(constants::phys::ElectronMass);
  }
  else {
    fit.SetParticleMass(frAlgo.GetDefaultParticleMass());
  }
  const int NHits = 3;  // triplets

  // prepare data
  int ista[NHits] = {fIstaL, fIstaM, fIstaR};

  ca::Station<fvec> sta[3];
  for (int is = 0; is < NHits; ++is) {
    sta[is] = frAlgo.fParameters.GetStation(ista[is]);
  };

  bool isMomentumFitted = ((staL().fieldStatus != 0) || (staM().fieldStatus != 0) || (staR().fieldStatus != 0));

  fvec ndfTrackModel = 4;                     // straight line
  ndfTrackModel += isMomentumFitted ? 1 : 0;  // track with momentum

  const fvec maxQp = frWData.CurrentIteration()->GetMaxQp();
  for (int i3 = 0; i3 < n3; ++i3) {

    // prepare data
    ca::HitIndex_t iwhit[NHits] = {fIhitL, fHitsM_3[i3], fHitsR_3[i3]};

    ca::HitIndex_t ihit[NHits] = {frWData.Hit(iwhit[0]).Id(), frWData.Hit(iwhit[1]).Id(), frWData.Hit(iwhit[2]).Id()};

    if (frAlgo.fParameters.DevIsMatchTripletsViaMc()) {
      int mc1 = frAlgo.GetMcTrackIdForCaHit(ihit[0]);
      int mc2 = frAlgo.GetMcTrackIdForCaHit(ihit[1]);
      int mc3 = frAlgo.GetMcTrackIdForCaHit(ihit[2]);
      if ((mc1 != mc2) || (mc1 != mc3)) {
        continue;
      }
    }

    fscal x[NHits], y[NHits], z[NHits], t[NHits];
    fscal dt2[NHits];
    ca::MeasurementXy<fvec> mxy[NHits];

    for (int ih = 0; ih < NHits; ++ih) {
      const ca::Hit& hit = frAlgo.fInputData.GetHit(ihit[ih]);
      mxy[ih] = ca::MeasurementXy<fvec>(hit.X(), hit.Y(), hit.dX2(), hit.dY2(), hit.dXY(), fvec::One(), fvec::One());

      x[ih]   = hit.X();
      y[ih]   = hit.Y();
      z[ih]   = hit.Z();
      t[ih]   = hit.T();
      dt2[ih] = hit.dT2();
    };

    // find the field along the track

    ca::FieldValue<fvec> B[3] _fvecalignment;
    ca::FieldRegion<fvec> fld _fvecalignment;
    ca::FieldRegion<fvec> fldTarget _fvecalignment;

    fvec tx[3] = {(x[1] - x[0]) / (z[1] - z[0]), (x[2] - x[0]) / (z[2] - z[0]), (x[2] - x[1]) / (z[2] - z[1])};
    fvec ty[3] = {(y[1] - y[0]) / (z[1] - z[0]), (y[2] - y[0]) / (z[2] - z[0]), (y[2] - y[1]) / (z[2] - z[1])};
    for (int ih = 0; ih < NHits; ++ih) {
      fvec dz = (sta[ih].fZ - z[ih]);
      B[ih]   = sta[ih].fieldSlice.GetFieldValue(x[ih] + tx[ih] * dz, y[ih] + ty[ih] * dz);
    };

    fld.Set(B[0], sta[0].fZ, B[1], sta[1].fZ, B[2], sta[2].fZ);
    fldTarget.Set(frWData.TargB(), frAlgo.GetParameters().GetTargetPositionZ(), B[0], sta[0].fZ, B[1], sta[1].fZ);

    TrackParamV& T = fit.Tr();

    // initial parameters
    {
      fvec dz01 = 1. / (z[1] - z[0]);
      T.Tx()    = (x[1] - x[0]) * dz01;
      T.Ty()    = (y[1] - y[0]) * dz01;
      T.Qp()    = 0.;
      T.Vi()    = 0.;
    }

    // repeat several times in order to improve the precision
    for (int iiter = 0; iiter < nIterations; ++iiter) {
      // fit forward
      {
        fit.SetQp0(T.Qp());
        fit.Qp0()(fit.Qp0() > maxQp)  = maxQp;
        fit.Qp0()(fit.Qp0() < -maxQp) = -maxQp;

        T.Qp()   = 0.;
        T.Vi()   = 0.;
        int ih0  = 0;
        T.X()    = x[ih0];
        T.Y()    = y[ih0];
        T.Z()    = z[ih0];
        T.Time() = t[ih0];

        T.ResetErrors(1., 1., 1., 1., 100., (sta[ih0].timeInfo ? dt2[ih0] : 1.e6), 1.e2);
        T.C00() = mxy[ih0].Dx2();
        T.C10() = mxy[ih0].Dxy();
        T.C11() = mxy[ih0].Dy2();

        T.Ndf()     = -ndfTrackModel + 2;
        T.NdfTime() = sta[ih0].timeInfo ? 0 : -1;

        //  add the target constraint
        fit.FilterWithTargetAtLine(frAlgo.GetParameters().GetTargetPositionZ(), frWData.TargetMeasurement(), fldTarget);

        for (int ih = 1; ih < NHits; ++ih) {
          fit.Extrapolate(z[ih], fld);
          auto radThick = frAlgo.fParameters.GetMaterialThickness(ista[ih], T.X(), T.Y());
          fit.MultipleScattering(radThick);
          fit.EnergyLossCorrection(radThick, fvec(-1.f));
          fit.FilterXY(mxy[ih]);
          fit.FilterTime(t[ih], dt2[ih], sta[ih].timeInfo);
        }
      }

      if (iiter == nIterations - 1) break;

      // fit backward
      {
        fit.SetQp0(T.Qp());
        fit.Qp0()(fit.Qp0() > maxQp)  = maxQp;
        fit.Qp0()(fit.Qp0() < -maxQp) = -maxQp;

        int ih0  = NHits - 1;
        T.X()    = x[ih0];
        T.Y()    = y[ih0];
        T.Z()    = z[ih0];
        T.Time() = t[ih0];
        T.Qp()   = 0.;
        T.Vi()   = 0.;

        T.ResetErrors(1., 1., 1., 1., 100., (sta[ih0].timeInfo ? dt2[ih0] : 1.e6), 1.e2);
        T.C00() = mxy[ih0].Dx2();
        T.C10() = mxy[ih0].Dxy();
        T.C11() = mxy[ih0].Dy2();

        T.Ndf()     = -ndfTrackModel + 2;
        T.NdfTime() = sta[ih0].timeInfo ? 0 : -1;

        for (int ih = NHits - 2; ih >= 0; --ih) {
          fit.Extrapolate(z[ih], fld);
          auto radThick = frAlgo.fParameters.GetMaterialThickness(ista[ih], T.X(), T.Y());
          fit.MultipleScattering(radThick);
          fit.EnergyLossCorrection(radThick, fvec(1.f));
          fit.FilterXY(mxy[ih]);
          fit.FilterTime(t[ih], dt2[ih], sta[ih].timeInfo);
        }
      }
    }  // for iiter

    fTracks_3[i3] = T;

    if constexpr (fDebugTriplets) {
      int ih0 = ihit[0];
      int ih1 = ihit[1];
      int ih2 = ihit[2];
      int mc1 = frAlgo.GetMcTrackIdForCaHit(ih0);
      int mc2 = frAlgo.GetMcTrackIdForCaHit(ih1);
      int mc3 = frAlgo.GetMcTrackIdForCaHit(ih2);
      if (1 || (mc1 >= 0) && (mc1 == mc2) && (mc1 == mc3)) {
        const ca::Hit& h0 = frAlgo.fInputData.GetHit(ih0);
        const ca::Hit& h1 = frAlgo.fInputData.GetHit(ih1);
        const ca::Hit& h2 = frAlgo.fInputData.GetHit(ih2);
        //const CbmL1MCTrack& mctr = CbmL1::Instance()->GetMcTracks()[mc1];
        LOG(info) << "== fitted triplet: "
                  << " iter " << frAlgo.fCurrentIterationIndex << " hits: {" << fIstaL << "/" << ih0 << " " << fIstaM
                  << "/" << ih1 << " " << fIstaR << "/" << ih2 << "} xyz: {" << h0.X() << " " << h0.Y() << " " << h0.Z()
                  << "}, {" << h1.X() << " " << h1.Y() << " " << h1.Z() << "}, {" << h2.X() << ", " << h2.Y() << ", "
                  << h2.Z() << "} chi2 " << T.GetChiSq()[0] << " ndf " << T.Ndf()[0] << " chi2time " << T.ChiSqTime()[0]
                  << " ndfTime " << T.NdfTime()[0];
        /*
        cbm::ca::tools::Debugger::Instance().FillNtuple(
          "triplets", mctr.iEvent, frAlgo.fCurrentIterationIndex, ih0, h0.X(), h0.Y(), h0.Z(), ih1, h1.X(), h1.Y(),
          h1.Z(), ih2, h2.X(), h2.Y(), h2.Z(), mc1, fIstaL, mctr.p, mctr.x, mctr.y, mctr.z, (float) T.GetChiSq()[0],
          (float) T.Ndf()[0], (float) T.ChiSqTime()[0], (float) T.NdfTime()[0]);
        */
      }
    }
  }  //i3

}  // FitTriplets


void TripletConstructor::StoreTriplets()
{
  /// Selects good triplets and saves them into fTriplets.
  /// Finds neighbouring triplets at the next station.

  Tindex n3 = fHitsM_3.size();

  bool isMomentumFitted = ((staL().fieldStatus != 0) || (staM().fieldStatus != 0) || (staR().fieldStatus != 0));

  fTriplets.clear();
  fTriplets.reserve(n3);

  const float tripletFinalChi2Cut = frWData.CurrentIteration()->GetTripletFinalChi2Cut();
  for (Tindex i3 = 0; i3 < n3; ++i3) {

    TrackParamV& T3 = fTracks_3[i3];

    // TODO: SG: normalize chi2, separate cuts on time and space

    fscal chi2 = T3.GetChiSq()[0] + T3.GetChiSqTime()[0];

    const ca::HitIndex_t ihitl = fIhitL;
    const ca::HitIndex_t ihitm = fHitsM_3[i3];
    const ca::HitIndex_t ihitr = fHitsR_3[i3];

    CBMCA_DEBUG_ASSERT(ihitl < frWData.HitStartIndexOnStation(fIstaL + 1));
    CBMCA_DEBUG_ASSERT(ihitm < frWData.HitStartIndexOnStation(fIstaM + 1));
    CBMCA_DEBUG_ASSERT(ihitr < frWData.HitStartIndexOnStation(fIstaR + 1));

    if (!frWData.CurrentIteration()->GetTrackFromTripletsFlag()) {
      if (chi2 > tripletFinalChi2Cut) {
        continue;
      }
    }

    // assert(std::isfinite(chi2) && chi2 > 0);

    if (!std::isfinite(chi2) || chi2 < 0) {
      continue;
    }

    //if (!T3.IsEntryConsistent(true, 0)) { continue; }

    fscal qp  = T3.Qp()[0];
    fscal Cqp = T3.C44()[0];

    // TODO: SG: a magic correction that comes from the legacy code
    // removing it leads to a higher ghost ratio
    Cqp += 0.001;

    fTriplets.emplace_back(ihitl, ihitm, ihitr, fIstaL, fIstaM, fIstaR, 0, 0, 0, chi2, qp, Cqp, T3.Tx()[0], T3.C22()[0],
                           T3.Ty()[0], T3.C33()[0], isMomentumFitted);
  }
}


void TripletConstructor::CollectHits(const TrackParamV& Tr, const int iSta, const double chi2Cut, const int iMC,
                                     Vector<ca::HitIndex_t>& collectedHits, int maxNhits)
{
  /// Collect hits on a station

  collectedHits.clear();
  collectedHits.reserve(maxNhits);

  const ca::Station<fvec>& sta = frAlgo.fParameters.GetStation(iSta);

  fFit.SetTrack(Tr);
  TrackParamV& T = fFit.Tr();
  //LOG(info) << T.chi2[0] ;
  T.ChiSq() = 0.;
  // if make it bigger the found hits will be rejected later because of the chi2 cut.
  const fvec Pick_m22 = (fvec(chi2Cut) - T.GetChiSq());

  const fscal timeError2 = T.C55()[0];
  const fscal time       = T.Time()[0];

  const auto& grid = frWData.Grid(iSta);
  const fvec maxDZ = frWData.CurrentIteration()->GetMaxDZ();
  ca::GridArea area(grid, T.X()[0], T.Y()[0], (sqrt(Pick_m22 * T.C00()) + grid.GetMaxRangeX() + maxDZ * abs(T.Tx()))[0],
                    (sqrt(Pick_m22 * T.C11()) + grid.GetMaxRangeY() + maxDZ * abs(T.Ty()))[0]);
  if constexpr (fDebugCollectHits) {
    LOG(info) << "search area: " << T.X()[0] << " " << T.Y()[0] << " "
              << (sqrt(Pick_m22 * T.C00()) + grid.GetMaxRangeX() + maxDZ * abs(T.Tx()))[0] << " "
              << (sqrt(Pick_m22 * T.C11()) + grid.GetMaxRangeY() + maxDZ * abs(T.Ty()))[0];
  }
  if (frAlgo.fParameters.DevIsIgnoreHitSearchAreas()) {
    area.DoLoopOverEntireGrid();
  }

  ca::HitIndex_t ih = 0;

  while (area.GetNextObjectId(ih) && ((int) collectedHits.size() < maxNhits)) {  // loop over station hits


    if (frWData.IsHitSuppressed(ih)) {
      continue;
    }

    const ca::Hit& hit = frWData.Hit(ih);
    if constexpr (fDebugCollectHits) {
      LOG(info) << "hit in the search area: " << hit.ToString();
      LOG(info) << "   check the hit.. ";
    }
    if (iMC >= 0) {  // match via MC
      if (iMC != frAlgo.GetMcTrackIdForWindowHit(ih)) {
        if constexpr (fDebugCollectHits) {
          LOG(info) << "   hit mc does not match";
        }
        continue;
      }
    }

    // check time-boundaries

    //TODO: move hardcoded cuts to parameters
    if ((sta.timeInfo) && (T.NdfTime()[0] >= 0)) {
      if (fabs(time - hit.T()) > 1.4 * (3.5 * sqrt(timeError2) + hit.RangeT())) {
        if constexpr (fDebugCollectHits) {
          LOG(info) << "   hit time does not match";
        }
        continue;
      }
      // if (fabs(time - hit.T()) > 30) continue;
    }

    // - check whether hit belong to the window ( track position +\- errors ) -
    const fscal z = hit.Z();

    // check y-boundaries
    auto [y, C11] = fFit.ExtrapolateLineYdY2(z);

    fscal dy_est = sqrt(Pick_m22[0] * C11[0]) + hit.RangeY();

    fscal xm = hit.X();
    fscal ym = hit.Y();

    const fscal dY = ym - y[0];

    if (fabs(dY) > dy_est) {
      if constexpr (fDebugCollectHits) {
        LOG(info) << "   hit Y does not match";
      }
      continue;
    }

    // check x-boundaries
    auto [x, C00] = fFit.ExtrapolateLineXdX2(z);

    fscal dx_est = sqrt(Pick_m22[0] * C00[0]) + hit.RangeX();

    const fscal dX = xm - x[0];

    if (fabs(dX) > dx_est) {
      if constexpr (fDebugCollectHits) {
        LOG(info) << "   hit X does not match";
      }
      continue;
    }

    // check chi2

    fvec C10 = fFit.ExtrapolateLineDxy(z);

    ca::MeasurementXy<fvec> mxy(hit.X(), hit.Y(), hit.dX2(), hit.dY2(), hit.dXY(), fvec::One(), fvec::One());

    auto [chi2x, chi2u] = ca::TrackFit::GetChi2XChi2U(mxy, x, y, C00, C10, C11);

    // TODO: adjust the cut, cut on chi2x & chi2u separately
    if (!frWData.CurrentIteration()->GetTrackFromTripletsFlag()) {
      if (chi2x[0] > chi2Cut) {
        if constexpr (fDebugCollectHits) {
          LOG(info) << "   hit chi2X is too large";
        }
        continue;
      }
      if (chi2u[0] > chi2Cut) {
        if constexpr (fDebugCollectHits) {
          LOG(info) << "   hit chi2U is too large";
        }
        continue;
      }
    }
    if constexpr (fDebugCollectHits) {
      LOG(info) << "   hit passed all the checks";
    }
    collectedHits.push_back(ih);

  }  // loop over the hits in the area
}
