/* Copyright (C) 2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CaOutputQa.h
/// \date   20.11.2023
/// \brief  A QA module for CA tracking (header)
/// \author S.Zharko <s.zharko@gsi.de>

#pragma once

#include "CaQaBase.h"

namespace cbm::algo::ca
{
  /// \class cbm::algo::ca::qa::OutputQa
  /// \brief OutputQa class for the CA tracking QA (header)
  ///
  class OutputQa : public QaBase {
   public:
    /// \brief Default destructor
    /// \param pSender  Pointer to the histogram sender
    OutputQa(std::shared_ptr<HistogramSender> pSender) : QaBase(pSender, "CA/Output"){};

    /// \brief Constructor from the configuration object
    /// \param config  QA configuration object
    OutputQa() = default;

    /// \brief Copy constructor
    OutputQa(const OutputQa&) = delete;

    /// \brief Move constructor
    OutputQa(OutputQa&&) = delete;

    /// \brief Destructor
    ~OutputQa() = default;

    /// \brief Copy assignment operator
    OutputQa& operator=(const OutputQa&) = delete;

    /// \brief Move assignment operator
    OutputQa& operator=(OutputQa&&) = delete;

    /// \brief QA execution function
    void Exec();

    /// \brief Initializes the QA
    void Init();

   private:
    static constexpr int knTrkParPoints = 2;  ///< Number of track points to build par distributions

    // Track distributions
    std::array<qa::H1D*, knTrkParPoints> fvphTrkTheta    = {{0}};    ///< hist: theta at first/last hit
    std::array<qa::H1D*, knTrkParPoints> fvphTrkPhi      = {{0}};    ///< hist: phi at first/last hit
    std::array<qa::H1D*, knTrkParPoints> fvphTrkChi2Ndf  = {{0}};    ///< hist: chi2/NDF at first/last hit
    std::array<qa::H1D*, knTrkParPoints> fvphHitSta      = {{0}};    ///< hist: station of first/last hit
    std::array<qa::H2D*, knTrkParPoints> fvphTrkPhiTheta = {{0}};    ///< hist: theta vs. phi at first/last hit
    qa::H1D* fphTrkNofHits                               = nullptr;  ///< hist: number of hits in track
  };
}  // namespace cbm::algo::ca
