/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CaQaBase.h
/// \date   29.02.2024
/// \brief  Base class for tracking QA (header)
/// \author Sergei Zharko <s.zharko@gsi.de>

#pragma once

#include "CaHit.h"  // for HitIndex_t
#include "CaTimesliceHeader.h"
#include "CaVector.h"
#include "QaData.h"  // QA data

namespace cbm::algo
{
  namespace qa
  {
    class H1D;
    class H2D;
  }  // namespace qa
  namespace ca
  {
    template<typename DataT>
    class Parameters;
    class InputData;
    class Track;
  }  // namespace ca
}  // namespace cbm::algo

namespace cbm::algo::ca
{
  /// \class cbm::algo::ca::qa::QaBase
  /// \brief Base class for CA tracking QA
  ///
  /// The class contains set of pointers to data structures, which are needed to build the histograms.
  class QaBase {
   public:
    /// \brief Default destructor
    /// \param pSender  Pointer to the histogram sender
    /// \param dirname  Directory name in the output file
    QaBase(std::shared_ptr<HistogramSender> pSender, const std::string& dirname) : fQaData(dirname), fpSender(pSender)
    {
    }

    /// \brief Constructor from the configuration object
    /// \param config  QA configuration object
    QaBase() = default;

    /// \brief Copy constructor
    QaBase(const QaBase&) = delete;

    /// \brief Move constructor
    QaBase(QaBase&&) = delete;

    /// \brief Destructor
    ~QaBase() = default;

    /// \brief Copy assignment operator
    QaBase& operator=(const QaBase&) = delete;

    /// \brief Move assignment operator
    QaBase& operator=(QaBase&&) = delete;

    /// \brief Check initialization
    /// \return true   All variables are initialized
    /// \return false  Some of are not initialized
    bool CheckInit() const;

    /// \brief Checks, if the histogram sender is defined
    bool IsSenderDefined() const { return static_cast<bool>(fpSender.get()); }

    /// \brief Registers tracking input data object
    /// \note  Call per TS
    void RegisterInputData(const InputData* pInputData) { fpInputData = pInputData; }

    /// \brief Registers track vector
    /// \note  Call per TS
    void RegisterTracks(const Vector<Track>* pvTracks) { fpvTracks = pvTracks; }

    /// \brief Registers reco hits indices vector
    /// \note  Call per TS
    void RegisterRecoHitIndices(const Vector<HitIndex_t>* pvRecoHits) { fpvRecoHits = pvRecoHits; }

    /// \brief Registers tracking parameters object
    /// \note  Call per run
    void RegisterParameters(const Parameters<fvec>* pParameters) { fpParameters = pParameters; }

   protected:
    qa::Data fQaData;  ///< QA data

    std::shared_ptr<HistogramSender> fpSender = nullptr;  ///< Histogram sender
    const Parameters<fvec>* fpParameters      = nullptr;  ///< Pointer to tracking parameters
    const InputData* fpInputData              = nullptr;  ///< Pointer to input data
    const Vector<Track>* fpvTracks            = nullptr;  ///< Pointer to tracks vector
    const Vector<HitIndex_t>* fpvRecoHits     = nullptr;  ///< Pointer to reco hit indices
  };
}  // namespace cbm::algo::ca
